"""Utilities for dealing with rotations, in 3D by default."""
import pandas as pd
import numpy as np


def random_rotation_matrix(shape=(1,)):
    """Fill an array of size shape with uniformly randomly distribution 3D
    rotation matrices, to get a final output of size (shape + (3,3)). Use James
    Arvo's fast method of generating a rotation about the z axis, then "moving
    the north pole" to a random point on the sphere through two conveniently
    fast (householder) reflections to uniformly sample SO(3)."""
    pass

def random_versor(shape=(1,)):
    """Use the 4 normal random numbers, normalized, to get a uniformly randomly
    distributed versor, isomorphic to an element of SO(3)."""
    pass

def versor_to_matrix(v):
    pass

def matrix_to_versor(m):
    pass



"""
Put the variables you want to load/save in a variable called
"variables_to_load"/"variables_to_save" then exec either
load_pickled_variables/save_pickled_variables.
"""

# load_pickle_fmt = '''
# try:
#     with open('{to_load:s}.pkl', 'rb') as f:
#         {to_load:s} = pickle.load(f)
# except FileNotFoundError:
#     pass
# '''
save_pickle_fmt = '''
try:
    {to_save:s}
    with open('{to_save:s}.pkl', 'wb') as f:
        pickle.dump({to_save:s}, f)
except:
    pass
'''

load_pickled_variables = """
import pickle
for var in variables_to_load:
    exec('''
try:
    with open('{to_load:s}.pkl', 'rb') as f:
        {to_load:s} = pickle.load(f)
except FileNotFoundError:
    pass
'''.format(to_load=var))
"""
save_pickled_variables = """
import pickle
for var in variables_to_save:
    exec('''
try:
    {to_save:s}
    with open('{to_save:s}.pkl', 'wb') as f:
        pickle.dump({to_save:s}, f)
except:
    pass
'''.format(to_save=var))
"""



import pandas as pd

def multiindex_col_ix(df, col):
    """Get the index into df.index.levels[] corresponding to a given "column
    name" in a multiindex."""
    return df.index.names.index(col)

def unique_vals_multiindex(df, col):
    """Get the "unique column values" in a multiindex "column"."""
    return df.index.levels[multiindex_col_ix(df, col)]

def fill_in_row_value(row, updater, col=None):
    """Function for apply'ing to rows (axis=1) that makes a new column by
    taking the correct value from a series whose index names should correspond
    to columns in the row. Optionally provide a default col to insert values
    from the existing row, otherwise insert NaN if updater series doesn't
    contain the desired key.

    Example
    -------
    TODO: make this example general
    >>> mmt['meiosis_time_combined'] = mmt.apply(bruno_util.pandas_util.fill_in_row_value,
    >>>         axis=1, updater=manual_meiosis_time, col='meiosis_time_guess')

    """
    cols = tuple(updater.index.names)
    if col and ~pd.isnull(row[col]):
        return row[col]
    if row[cols] in updater:
        return updater[row[cols]]
    return np.nan

def df_update_using_cols_like_keys(df1, df2, cols, ignore_index=True):
    """Add the rows in df1 whose values in cols do not appear in df2 to df2.

    Loosely, update df1 with the values of df2 as if cols were the index,
    but treating each block of values in cols as a whole.

    WARNING: only works if df1/df2 have "no" index for now, i.e. ignore_index
    (as in original index) must be True.

    Performed by taking the multiindex dfN.set_index(cols), taking the set
    difference, then appending to a copy of df2 by indexing back into df1 with
    the result."""
    df_diff = df_dict_diff_cols_like_keys(df1, df2, cols, ignore_index)
    return df_diff.append(df2)

def df_dict_diff_cols_like_keys(df1, df2, cols, ignore_index=True):
    """Return 'dict' difference df1 \ df2 using values in cols as the 'keys',
    but treat the non-unique keys as a single unit each."""
    if not ignore_index:
        raise NotImplementedError('only ignore_index=True is allowed.')
    df1_indexed = df1.set_index(cols)
    df2_indexed = df2.set_index(cols)
    diff_index = df1_indexed.index.difference(df2_indexed.index)

    new_df = [df1_indexed.loc[[idx]].reset_index() for idx in diff_index]
    # explicitly include initial object of append in case new_df ends up empty
    new_df.append(pd.DataFrame(index=[], columns=df2.columns))
    return pd.concat(new_df, ignore_index=True)






import collections
import itertools
import multiprocessing

class SimpleMapReduce(object):

    def __init__(self, map_func, reduce_func, num_workers=None,
                 pool_type='process'):
        """
        map_func

          Function to map inputs to intermediate data. Takes as argument
          one input source and returns a list of objects that can be
          unpacked as a (key, value) pair to be reduced.

        reduce_func

          Function to reduce partitioned version of intermediate data to
          final output. Takes a single argument item=(key,valuelist)
          where key is as produced by map_func and valuelist is the
          sequence of the values associated with that key, one value for
          each worker that produced that key.

        num_workers

          The number of workers to create in the pool. Defaults to the
          number of CPUs available on the current host.
        """
        self.map_func = map_func
        self.reduce_func = reduce_func
        if pool_type == 'thread':
            self.pool = multiprocessing.pool.ThreadPool(num_workers)
        elif pool_type == 'process':
            self.pool = multiprocessing.Pool(num_workers)
        else:
            raise NameError('No such multiprocessing pool type: ' +
                            pool_type)

    def partition(self, mapped_values):
        """Organize the mapped values by their key.
        Returns an unsorted sequence of tuples with a key and a sequence of values.
        """
        partitioned_data = collections.defaultdict(list)
        for key, value in mapped_values:
            partitioned_data[key].append(value)
        return partitioned_data.items()

    def __call__(self, inputs, chunksize=1):
        """Process the inputs through the map and reduce functions given.

        inputs
          An iterable containing the input data to be processed.

        chunksize=1
          The portion of the input data to hand to each worker.  This
          can be used to tune performance during the mapping phase.
        """
        # get list of responses from each worker, which should each be lists of
        # tuples they mapped
        map_responses = self.pool.map(self.map_func, inputs, chunksize=chunksize)
        # combine the list of tuples from all workers into a master list, then
        # partition it based on key
        partitioned_data = self.partition(itertools.chain(*map_responses))
        # pass the .items() iterator with the promised key,valuelist pairs
        # to the reduce_func
        reduced_values = self.pool.map(self.reduce_func, partitioned_data)
        return reduced_values



import numpy as np


def center_by_mass(x, particle_axis=0):
    """Subtract center of mass (unweighted) from a collection of vectors
    corresponding to particles's coordinates."""
    shape = x.shape
    # center of mass is the average position of the particles, so average over
    # the particles
    centers_of_mass = np.mean(x, particle_axis)
    # tile the centers of mass for clean elementwise division
    # is there a way to do this faster by relying on numpy's array projection
    # semantics?
    tile_shape = [shape[i] if i == particle_axis else 1
                  for i in range(len(shape))]
    centers_of_mass = np.tile(centers_of_mass, tile_shape)
    return x - centers_of_mass

def locally_linear_fit(x, y, window_size=5, **kwargs):
    """Plots the local slope of y(x) using a moving average with the given
    window size. Passes kwargs onto ax.plot."""
    if window_size % 2 == 0:
        raise ValueError('window_size must be odd')
    if window_size < 3:
        raise ValueError('window_size must be at least 3, so that we have at'
                         ' least two points to fit at endpoints of array.')
    lenx = len(x)
    if lenx < 2:
        raise ValueError('Can\'t fit less than 2 points!')
    inc = int(window_size / 2)
    slope = np.zeros_like(x)
    for i in range(lenx):
        imin = np.max([0, i - inc])
        imax = np.min([lenx, i + inc + 1])
        xfit = x[imin:imax]
        yfit = y[imin:imax]
        to_fit = np.isfinite(xfit) & np.isfinite(yfit)
        if not np.any(to_fit):
            slope[i] = np.nan
        else:
            slope[i], intcpt, r_val, p_val, std_err = stats.linregress(xfit[to_fit], yfit[to_fit])
    ax = kwargs.pop('axes', None)
    if ax is not None:
        ax.plot(x, slope, **kwargs)
    return ax, slope



import os

def path_parts(path):
    folders = []
    while True:
        path, folder = os.path.split(path)
        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)
            break
    folders.reverse()

def get_unique_folder(basename):
    num = 0
    while True:
        try:
            folder = basename + '.' + str(num)
            os.makedirs(folder, mode=0o755)
            break
        except OSError:
            num = num + 1
    return folder



"""A module for stuff that should be part of Python itself by default but
isn't."""


def make_decorator_factory_args_optional(decorator):
    """Decorator for decorator (factories) that allows them to be used as
    normal decorators. In other words, if you have a decorator that takes
    arguments, then
        @make_decorator_factory_args_optional
        def decorator(arg1=blah, **kwargs)
    will allow you to use the decorator as
        @decorator
        def func():
    instead of only as
        @decorator()
        def func():
    so that you don't have to tell your users to stick with just one.

    WARNING: This make_decorator_factory_args_optional only works if the
    decorator that it's decorating cannot ever be called with a single,
    non-keyword, callable argument in factory mode. i.e. if your decorator can
    be called like
        @decorator(lambda x: x*x)
        def func():
    then DO NOT USE make_decorator_factory_args_optional!"""
    def wrapped_decorator(*args, **kwargs):
        if len(args) == 1 and callable(args[0]) and len(kwargs) == 0:
            return decorator()(args[0])
        else:
            return decorator(*args, **kwargs)
    return wrapped_decorator

@make_decorator_factory_args_optional
def well_behaved_decorator(has_params=False, single_callable_param=False):
    """Decorator for decorators to allow them to preserve docstrings, __name__
    and __dict__ of wrapped functions, so long as the decorator is fairly
    simple. Helpful for documentation purposes.

    If your decorator takes arguments, you'll want to specify this via
    has_params.

    If a decorator doesn't modify function attributes or docstring, and if it
    doesn't take a single, non-keyword-only, callable argument, then it is
    eligible to use this.

    Whenever a decorator is passed a single, callable, non-keyword-only
    parameter, it is likely that the decorator is being called in "wrapper" or
    normal mode, as
        @decorate
        def func():
    If your decorator takes arguments, then you would normally have to call it
    in "factory" mode, as
        @decorate()
        def func():
    You poor users likely will be miffed that they have to remember which one
    to use, but make_decorator_factory_args_optional can allow decorators
    that take optional arguments to be used in the "normal" mode. However, the
    only way to tell if a decorator is being called in the former mode is to
    check if its argument list consists of a single, non-keyword, callable
    argument. We use the same criterion here to correctly extract the identity
    of the function to be wrapped.

    If your decorator does take a single, callable, non-keyword-only parameter
    (screw you), then you can pass single_callable_param=True to specify that
    you would always like your decorator to be treated as if it's in factory
    mode when we decide what the function being wrapped is. You gain back the
    ability to use well_behaved_decorator, but you lose the ability to use your
    decorator in "normal" mode, if you had somehow managed to make that work
    while having a single, positional, callable argument....
    """
    # a decorator with no arguments is just a function that takes a single
    # argument (the function to be decorated) and returns the decorated version
    # of the function
    def wrap_no_params(decorator):
        """Test wrap_no_params"""
        def wrapped_decorator(f):
            """Test wrap_no_params.wrapped_decorator"""
            # gives the wrapped functions properties of the original functions
            g = decorator(f)
            g.__name__ = f.__name__
            g.__doc__ = f.__doc__
            g.__dict__.update(f.__dict__)
            return g
        # gives the wrapped decorator properties of the original decorator
        wrapped_decorator.__name__ = decorator.__name__
        wrapped_decorator.__doc__ = decorator.__doc__
        wrapped_decorator.__dict__.update(decorator.__dict__)
        return wrapped_decorator

    # a decorator with parameters is just a function that takes arguments then
    # returns a decorator with no arguments
    def wrap_with_params(decorator):
        """Factory that returns the decorated decorator if that decorator takes
        arguments. Due to Python's nonsense, we have to manually check if the
        decorator to be decorated is being called with a single callable as its
        argument, as this is our only indication that it is in fact being
        called in normal decorator mode, e.g.
        @decorator
        def func(): ...
        instead of in factory mode, e.g.
        @decorator()
        def func(): ...

        This is because in the former case, we must return the decorated func,
        and in the latter case, we must return a decorator that can decorate
        func.

        Because of this, if you make a decorator that takes a single, callable,
        non-kw argument, then screw you.
        """
        def wrapped_decorator(*args, **kwargs):
            """Centralizes the check to see if the decorator was called in
            factory mode."""
            # the decoration was made in regular mode, so no params are
            # actually used even if the decorator can take them
            if len(args) == 1 and callable(args[0]) and len(kwargs) == 0 \
                    and not single_callable_param:
                return wrap_no_params(decorator)(args[0])
            # otherwise the decoration was made in factory mode, so we must
            # call the decorator itself to get a decorator with no args back
            decorator_given_params = decorator(*args, **kwargs)
            return wrap_no_params(decorator_given_params)
        return wrapped_decorator

    # we are a decorator with params (i.e. a decorator factory), so we return a
    # function that takes a single argument, the decorator to be decorated :)
    if has_params:
        return wrap_with_params
    else:
        return wrap_no_params



import multiprocessing
import string

from multiprocessing_mapreduce import SimpleMapReduce

def file_to_words(filename):
    """Read a file and return a sequence of (word, occurances) values.
    """
    STOP_WORDS = set([
            'a', 'an', 'and', 'are', 'as', 'be', 'by', 'for', 'if', 'in',
            'is', 'it', 'of', 'or', 'py', 'rst', 'that', 'the', 'to', 'with',
            ])
    TR = string.maketrans(string.punctuation, ' ' * len(string.punctuation))

    print(multiprocessing.current_process().name, 'reading', filename)
    output = []

    with open(filename, 'rt') as f:
        for line in f:
            if line.lstrip().startswith('..'): # Skip rst comment lines
                continue
            line = line.translate(TR) # Strip punctuation
            for word in line.split():
                word = word.lower()
                if word.isalpha() and word not in STOP_WORDS:
                    output.append( (word, 1) )
    return output


def count_words(item):
    """Convert the partitioned data for a word to a
    tuple containing the word and the number of occurances.
    """
    word, occurances = item
    return (word, sum(occurances))


if __name__ == '__main__':
    import operator
    import glob

    input_files = glob.glob('../*.py')

    mapper = SimpleMapReduce(file_to_words, count_words)
    word_counts = mapper(input_files)
    word_counts.sort(key=operator.itemgetter(1))
    word_counts.reverse()

    print '\nTOP 20 WORDS BY FREQUENCY\n'
    top20 = word_counts[:20]
    longest = max(len(word) for word, count in top20)
    for word, count in top20:
        print '%-*s: %5s' % (longest+1, word, count)



import os
import struct
import numpy as np

def strong_default_seed(f):
    """Use os.random to correctly seed a np.random.RandomState-compatible
    random variate if the user does not provide the "seed" random variable. The
    correctly initialized state is passed to the function via the "random_state"
    kwarg.

    Parameters
    ----------
    f : Callable
        function that takes at least two kwargs, seed and random_state. random_state should
        be used to generate random numbers within the function.

    Returns
    -------
    wrapped_f : Callable
        Function that initializes random_state (using seed, or by generating a seed if
        none is given) if it is not passed to the function.

    Notes
    -----
    On sane operating systems, you should be able to parallelize the function
    and get different results for each invocation by letting this wrapper
    initiazlie the seed at each call site.
    """
    def wrapped_f(*args, **kwargs):
        if 'random_state' in kwargs and kwargs['random_state'] is not None:
            return f(*args, **kwargs)
        if 'seed' not in kwargs or kwargs['seed'] is None:
            seed = struct.unpack("<L", os.urandom(4))[0]
        else:
            seed = kwargs['seed']
        random_state = np.random.RandomState()
        random_state.seed(seed)
        kwargs['random_state'] = seed
        return f(*args, **kwargs)
    return wrapped_f

def make_pool(rvs_f, pool_size=1024, unit_size=tuple(), **kwargs):
    """Use a random variate generator that takes a "size" keyword to make a
    function that just return random values "one" at a time from a pre-computed
    pool, replenishing the pool as necessary.

    This often provides serious speedups over calling the function with
    size=unit_size repeatedly when usign numpy/scipy random generation
    functions.

    kwargs are passed on to the rvs_f function when it is called to replenish
    the pool.

    Parameters
    ----------

    rvs_f : Callable
        function that takes size kwarg and generates that many random numbers
    pool_size : int
        how many sets of random numbers to generate ahead of time. this is how
        many times the function we generate can be called before replenishing
        its pool.
    unit_size : Tuple[int]
        the shape of the set of random numbers to be returned every time the
        wrapped function is called

    Returns
    -------

    wrapped_rvs_f : Callable
        function that returns random array of shape unit_size every time it is
        called (ignores all arguments) from a precomputed pool.
    """
    def rv():
        if rv.ri == pool_size:
            rv.rand_pool = rvs_f(size=((pool_size, )+unit_size), **kwargs)
            rv.ri = 0
        rv.ri += 1
        return rv.rand_pool[rv.ri - 1]
    rv.rand_pool = rvs_f(size=((pool_size,)+unit_size), **kwargs)
    rv.ri = 0
    return rv




#!/usr/bin/env python

from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='bruno_util',
      version='1.3.1',
      description='Catch-all package for utilities useful to Bruno Beltran',
      long_description=readme(),
      author='Bruno Beltran',
      author_email='brunobeltran0@gmail.com',
      packages=['bruno_util'],
      license='MIT',
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Topic :: Utilities'
      ],
      keywords='parameter scanning scientific search testing',
      url='https://github.com/brunobeltran/bruno_util',
      install_requires=['pscan'], # include other packages I wrote that are useful
     )
